/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client;

import de.xoev.xta.client.encyrption.EncryptionService;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.model.Attachment;
import de.xoev.xta.client.model.Content;
import de.xoev.xta.client.model.DeliveryOptions;
import de.xoev.xta.client.model.Identifier;
import de.xoev.xta.client.model.Message;
import de.xoev.xta.client.model.MessageFilter;
import de.xoev.xta.client.util.XmlUtil;
import de.xoev.transport.xta._211.ContentType;
import de.xoev.transport.xta._211.GenericContentContainer;
import de.xoev.transport.xta._211.GenericContentContainer.ContentContainer;
import de.xoev.transport.xta._211.LookupServiceRequest;
import de.xoev.transport.xta._211.LookupServiceType;
import eu.osci.ws._2008._05.transport.MsgBoxCloseRequestType;
import eu.osci.ws._2008._05.transport.MsgBoxFetchRequest;
import eu.osci.ws._2008._05.transport.MsgBoxStatusListRequestType;
import eu.osci.ws._2008._05.transport.MsgSelector;
import eu.osci.ws._2014._10.transport.DeliveryAttributesType;
import eu.osci.ws._2014._10.transport.DestinationsType;
import eu.osci.ws._2014._10.transport.KeyCodeType;
import eu.osci.ws._2014._10.transport.MessageMetaData;
import eu.osci.ws._2014._10.transport.MsgIdentificationType;
import eu.osci.ws._2014._10.transport.OriginatorsType;
import eu.osci.ws._2014._10.transport.PartyIdentifierType;
import eu.osci.ws._2014._10.transport.PartyType;
import eu.osci.ws._2014._10.transport.QualifierType;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.w3._2005._08.addressing.AttributedURIType;

/**
 * Basis-Klasse des XTA-Client
 */
@Slf4j
abstract class XtaBaseClient {

  private final static eu.osci.ws._2014._10.transport.ObjectFactory objectFactoryOcsi2 = new eu.osci.ws._2014._10.transport.ObjectFactory();

  private final static eu.osci.ws._2008._05.transport.ObjectFactory objectFactoryOcsi1 = new eu.osci.ws._2008._05.transport.ObjectFactory();

  private final static de.xoev.transport.xta._211.ObjectFactory objectFactoryXta = new de.xoev.transport.xta._211.ObjectFactory();

  private final static org.w3._2005._08.addressing.ObjectFactory objectFactoryWsAddressing = new org.w3._2005._08.addressing.ObjectFactory();

  static LookupServiceRequest createLookupServiceRequest(final DeliveryOptions deliveryOptions, final Identifier reader) {
    final LookupServiceRequest lookupServiceRequest = objectFactoryXta.createLookupServiceRequest();
    final LookupServiceRequest.LookupServiceRequestList lookupServiceRequestList = objectFactoryXta
        .createLookupServiceRequestLookupServiceRequestList();
    final LookupServiceType lookupServiceType = objectFactoryXta.createLookupServiceType();
    lookupServiceType.setServiceType(deliveryOptions.getService());
    lookupServiceType.setReader(createPartyType(reader));
    lookupServiceRequestList.setLookupService(lookupServiceType);
    lookupServiceRequest.getLookupServiceRequestList().add(lookupServiceRequestList);
    return lookupServiceRequest;
  }

  static ContentType createContentType(final Content content) {
    final ContentType contentType = objectFactoryXta.createContentType();
    contentType.setContentDescription(content.getContentDescription());
    contentType.setContentType(content.getContentType());
    contentType.setEncoding(content.getEncoding());
    contentType.setFilename(content.getFilename());
    contentType.setId(content.getId());
    contentType.setLang(content.getLanguage());
    contentType.setSize(content.getSize());
    contentType.setValue(content.getData());
    return contentType;
  }

  static GenericContentContainer createGenericContentContainer(final XtaClient xtaClient, final DeliveryOptions deliveryOptions,
      final Message message,
      final List<Attachment> attachmentList) throws ClientException {

    final GenericContentContainer genericContentContainer = objectFactoryXta.createGenericContentContainer();
    final GenericContentContainer.ContentContainer contentContainer = objectFactoryXta
        .createGenericContentContainerContentContainer();

    contentContainer.setMessage(createContentType(message));
    if (CollectionUtils.isNotEmpty(attachmentList)) {
      for (final Attachment attachment : attachmentList) {
        contentContainer.getAttachment().add(createContentType(attachment));
      }
    }

    if (StringUtils.isNotEmpty(deliveryOptions.getEncryptionServiceProfilename())) {
      final EncryptionService encryptionServiceProfile = xtaClient
          .getEncryptionService(deliveryOptions.getEncryptionServiceProfilename());
      genericContentContainer.setEncryptedData(encryptionServiceProfile.encrypt(contentContainer));
    } else {
      genericContentContainer.setContentContainer(contentContainer);
    }
    return genericContentContainer;
  }

  static PartyType createPartyType(final Identifier identifier) {
    PartyType partyType = objectFactoryOcsi2.createPartyType();
    PartyIdentifierType partyIdentifierType = objectFactoryOcsi2.createPartyIdentifierType();
    partyIdentifierType.setValue(identifier.getId());
    partyIdentifierType.setType(identifier.getType());
    partyIdentifierType.setName(identifier.getName());
    partyIdentifierType.setCategory(identifier.getCategory());
    partyType.setIdentifier(partyIdentifierType);
    return partyType;
  }

  static AttributedURIType createAttributedURIType(final String uri) {
    final AttributedURIType attributedURIType = objectFactoryWsAddressing.createAttributedURIType();
    attributedURIType.setValue(uri);
    return attributedURIType;
  }

  static MessageMetaData createMessageMetaData(final Identifier author, final Identifier reader,
      final AttributedURIType messageId, final DeliveryOptions deliveryOptions, final Message message) throws ClientException {

    final MessageMetaData messageMetaData = objectFactoryOcsi2.createMessageMetaData();
    messageMetaData.setDeliveryAttributes(createDeliveryAttributesType(deliveryOptions));
    messageMetaData.setOriginators(createOriginatorsType(author, deliveryOptions.getReplyTo()));
    messageMetaData.setDestinations(createDestinationsType(reader));
    messageMetaData.setMsgIdentification(createMsgIdentificationType(messageId));
    messageMetaData.setQualifier(createQualifierType(deliveryOptions, message));
    messageMetaData.setTestMsg(deliveryOptions.getIsTestMessage());

    return messageMetaData;
  }

  static DeliveryAttributesType createDeliveryAttributesType(final DeliveryOptions deliveryOptions) throws ClientException {

    final DeliveryAttributesType deliveryAttributesType = objectFactoryOcsi2.createDeliveryAttributesType();
    if (deliveryOptions.getOrigin() != null) {
      deliveryAttributesType.setOrigin(XmlUtil.toXmlGregorianCalender(deliveryOptions.getOrigin()));
    } else {
      deliveryAttributesType.setOrigin(XmlUtil.toXmlGregorianCalender(new Date()));
    }
    if (deliveryOptions.getSendNotBefore() != null) {
      deliveryAttributesType.setNotBefore(XmlUtil.toXmlGregorianCalender(deliveryOptions.getSendNotBefore()));
    }

    if (deliveryOptions.getObsoleteAfter() != null) {
      deliveryAttributesType.setObsoleteAfter(XmlUtil.toXmlGregorianCalender(deliveryOptions.getObsoleteAfter()));
    }

    if (StringUtils.isNotEmpty(deliveryOptions.getEncryptionServiceProfilename())) {
      deliveryAttributesType.setServiceQuality(deliveryOptions.getEncryptionServiceProfilename());
    }

    return deliveryAttributesType;
  }

  static OriginatorsType createOriginatorsType(final Identifier identifier, final Identifier replyTo) {
    final OriginatorsType originatorsType = objectFactoryOcsi2.createOriginatorsType();
    originatorsType.setAuthor(createPartyType(identifier));
    if (replyTo != null) {
      originatorsType.setReplyTo(createPartyType(replyTo));
    }
    return originatorsType;
  }

  static DestinationsType createDestinationsType(final Identifier identifier) {
    final DestinationsType destinationsType = objectFactoryOcsi2.createDestinationsType();
    destinationsType.setReader(createPartyType(identifier));
    return destinationsType;
  }

  static MsgIdentificationType createMsgIdentificationType(final AttributedURIType messageId) {
    final MsgIdentificationType msgIdentificationType = objectFactoryOcsi2.createMsgIdentificationType();
    msgIdentificationType.setMessageID(messageId);
    return msgIdentificationType;
  }

  static QualifierType createQualifierType(final DeliveryOptions deliveryOptions, final Message message) {
    final QualifierType qualifierType = objectFactoryOcsi2.createQualifierType();
    qualifierType.setService(deliveryOptions.getService());
    qualifierType.setBusinessScenario(createBusinessScenario(deliveryOptions));
    qualifierType.setMessageType(createMessageType(message));
    return qualifierType;
  }

  static QualifierType.MessageType createMessageType(final Message message) {
    final QualifierType.MessageType messageType = objectFactoryOcsi2.createQualifierTypeMessageType();
    messageType.setPayloadSchema(message.getMessageTypePayloadSchema());
    messageType.setCode(message.getMessageTypeCode());
    messageType.setListURI("urn:de:payloadSchema:elementName");
    messageType.setListVersionID("1.0");
    return messageType;
  }

  static QualifierType.BusinessScenario createBusinessScenario(final DeliveryOptions deliveryOptions) {
    final QualifierType.BusinessScenario businessScenario = objectFactoryOcsi2.createQualifierTypeBusinessScenario();
    if (deliveryOptions.isDefinedBusinessScenario()) {
      businessScenario.setDefined(createKeyCodeType(deliveryOptions.getBusinessScenario()));
    } else {
      businessScenario.setUndefined(deliveryOptions.getBusinessScenario());
    }
    return businessScenario;
  }

  static KeyCodeType createKeyCodeType(String businessScenario) {
    final KeyCodeType keyCodeType = objectFactoryOcsi2.createKeyCodeType();
    keyCodeType.setListURI("urn:de:xta:codeliste:business.scenario_1");
    keyCodeType.setListVersionID("1");
    keyCodeType.setCode(businessScenario);
    return keyCodeType;
  }

  static MsgBoxCloseRequestType createMsgBoxCloseRequestType(final String messageId, final String requestHandle) {
    final MsgBoxCloseRequestType type = objectFactoryOcsi1.createMsgBoxCloseRequestType();
    type.getLastMsgReceived().add(createAttributedURIType(messageId));
    type.setMsgBoxRequestID(requestHandle);
    return type;
  }

  static ContentContainer getContentContainer(XtaClient xtaClient, GenericContentContainer genericContentContainer,
      final String serviceProfileName) throws ClientException {
    if (genericContentContainer != null && genericContentContainer.getContentContainer() != null) {
      if (StringUtils.isNotEmpty(serviceProfileName)) {
        final EncryptionService service = xtaClient.getEncryptionService(serviceProfileName);
        return service.decyrpt(genericContentContainer.getEncryptedData());
      } else {
        return genericContentContainer.getContentContainer();
      }
    }
    return null;
  }

  static Message getMessage(final ContentContainer contentContainer) {
    if (contentContainer != null && contentContainer.getMessage() != null) {
      final ContentType contentType = contentContainer.getMessage();
      return new Message(contentType.getContentType(), contentType.getValue(), contentType.getContentDescription(),
          contentType.getEncoding(), contentType.getFilename(), contentType.getId(), contentType.getLang(),
          contentType.getSize());
    }

    return null;
  }

  static List<Attachment> getAttachments(final ContentContainer contentContainer) {
    final List<Attachment> attachments = new ArrayList<>();
    if (contentContainer != null) {
      final List<ContentType> contentTypeList = contentContainer.getAttachment();
      if (CollectionUtils.isNotEmpty(contentTypeList)) {
        for (final ContentType contentType : contentTypeList) {
          attachments.add(createAttachment(contentType));
        }
      }
    }
    return attachments;
  }

  static Attachment createAttachment(final ContentType contentType) {
    return new Attachment(contentType.getContentType(), contentType.getValue(), contentType.getContentDescription(),
        contentType.getEncoding(), contentType.getFilename(), contentType.getId(), contentType.getLang(), contentType.getSize());
  }

  static MsgBoxFetchRequest createMsgBoxFetchRequest(final String messageId) {
    final MsgBoxFetchRequest msgBoxFetchRequest = objectFactoryOcsi1.createMsgBoxFetchRequest();
    final MsgSelector msgSelector = objectFactoryOcsi1.createMsgSelector();
    msgBoxFetchRequest.setMsgSelector(msgSelector);
    msgSelector.getMessageID().add(createAttributedURIType(messageId));
    return msgBoxFetchRequest;
  }

  static MsgBoxStatusListRequestType createMsgBoxStatusListRequestType(final MessageFilter filter) throws ClientException {
    final MsgBoxStatusListRequestType type = objectFactoryOcsi1.createMsgBoxStatusListRequestType();

    type.setListForm("MessageMetaData");

    type.setMaxListItems(BigInteger.valueOf(filter.getCount()));
    final MsgSelector msgSelector = objectFactoryOcsi1.createMsgSelector();
    type.setMsgSelector(msgSelector);

    if (filter.getTimeFrom() != null) {
      msgSelector.setMsgBoxEntryTimeFrom(XmlUtil.toXmlGregorianCalender(filter.getTimeFrom()));
    }

    if (filter.getTimeTo() != null) {
      msgSelector.setMsgBoxEntryTimeTo(XmlUtil.toXmlGregorianCalender(filter.getTimeTo()));
    }

    if (filter.isNewEntries()) {
      msgSelector.setNewEntry(true);
    }

    if (ArrayUtils.isNotEmpty(filter.getMessageId())) {
      for (String messageId : filter.getMessageId()) {
        msgSelector.getMessageID().add(createAttributedURIType(messageId));
      }
    }

    return type;
  }
}
