/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client;

import com.google.common.base.Stopwatch;
import de.xoev.xta.client.core.ManagementServiceClient;
import de.xoev.xta.client.core.MsgBoxServiceClient;
import de.xoev.xta.client.core.SendServiceClient;
import de.xoev.xta.client.encyrption.EncryptionService;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.exception.ClientInitializationException;
import de.xoev.xta.client.model.Attachment;
import de.xoev.xta.client.model.DeliveryOptions;
import de.xoev.xta.client.model.GetMessageResult;
import de.xoev.xta.client.model.Identifier;
import de.xoev.xta.client.model.Message;
import de.xoev.xta.client.model.MessageFilter;
import de.xoev.xta.client.model.MessageTransportReport;
import de.xoev.transport.xta._211.CancelDeniedException;
import de.xoev.transport.xta._211.GenericContentContainer;
import de.xoev.transport.xta._211.GenericContentContainer.ContentContainer;
import de.xoev.transport.xta._211.InvalidMessageIDException;
import de.xoev.transport.xta._211.LookupServiceResponse;
import de.xoev.transport.xta._211.LookupServiceResultType;
import de.xoev.transport.xta._211.MessageSchemaViolationException;
import de.xoev.transport.xta._211.MessageVirusDetectionException;
import de.xoev.transport.xta._211.ParameterIsNotValidException;
import de.xoev.transport.xta._211.PermissionDeniedException;
import de.xoev.transport.xta._211.SyncAsyncException;
import de.xoev.transport.xta._211.TransportReport;
import de.xoev.transport.xta._211.XTAWSTechnicalProblemException;
import eu.osci.ws._2008._05.transport.MsgBoxResponseType;
import eu.osci.ws._2008._05.transport.MsgStatusListType;
import eu.osci.ws._2014._10.transport.MessageMetaData;
import eu.osci.ws._2014._10.transport.PartyType;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import javax.xml.ws.Holder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.w3._2005._08.addressing.AttributedURIType;

/**
 * XTA-Client zum Senden und Emfangen von Nachrichten.
 */
@Slf4j
public class XtaClient extends XtaBaseClient {

  private Identifier identity;
  private SendServiceClient sendServiceClient;
  private MsgBoxServiceClient msgBoxServiceClient;
  private ManagementServiceClient managementServiceClient;
  private EncryptionService[] encryptionServices;

  public XtaClient(final XtaClientConfig config) throws ClientInitializationException {

    this.identity = config.getAuthor();

    // ManagementPort initialisieren
    managementServiceClient = ManagementServiceClient.builder(config.getManagementServiceUrl())
        .setClientCertKeystore(config.getClientCertKeystore())
        .setTrustCertKeystore(config.getTrustCertKeystore())
        .setHttpProxy(config.getHttpProxy())
        .setSoapRequestLogging(config.isLogSoapRequest())
        .setSoapResponseLogging(config.isLogSoapResponse()).setValidation(config.isValidation())
        .setHttpClientPolicy(config.getHttpClientPolicy())
        .build();

    // SendPort initialisieren
    if (config.getSendServiceUrl() != null) {
      sendServiceClient = SendServiceClient.builder(config.getSendServiceUrl())
          .setClientCertKeystore(config.getClientCertKeystore())
          .setTrustCertKeystore(config.getTrustCertKeystore())
          .setHttpProxy(config.getHttpProxy()).setSoapRequestLogging(config.isLogSoapRequest())
          .setSoapResponseLogging(config.isLogSoapResponse()).setValidation(config.isValidation())
          .setHttpClientPolicy(config.getHttpClientPolicy())
          .build();
    }

    // MsgPort initialisieren
    if (config.getMsgBoxServiceUrl() != null) {
      msgBoxServiceClient = MsgBoxServiceClient.builder(config.getMsgBoxServiceUrl())
          .setClientCertKeystore(config.getClientCertKeystore())
          .setTrustCertKeystore(config.getTrustCertKeystore())
          .setHttpProxy(config.getHttpProxy()).setSoapRequestLogging(config.isLogSoapRequest())
          .setSoapResponseLogging(config.isLogSoapResponse()).setValidation(config.isValidation())
          .setHttpClientPolicy(config.getHttpClientPolicy())
          .build();
    }

    this.encryptionServices = config.getEncryptionServices();

  }

  /**
   * Senden einer Nachricht ohne Attachments an den XTA-Broker
   *
   * @param reader          Identität des Leser
   * @param deliveryOptions Sendeoptionen
   * @param message         Nachricht
   * @return
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   * @throws MessageSchemaViolationException
   * @throws ParameterIsNotValidException
   * @throws MessageVirusDetectionException
   * @throws SyncAsyncException
   * @throws InvalidMessageIDException
   * @throws ClientException
   */
  public MessageTransportReport sendMessage(final Identifier reader, final DeliveryOptions deliveryOptions, final Message message)
      throws ClientException, PermissionDeniedException, XTAWSTechnicalProblemException, MessageVirusDetectionException, MessageSchemaViolationException, ParameterIsNotValidException, SyncAsyncException, InvalidMessageIDException {

    return sendMessage(reader, deliveryOptions, message, new ArrayList<>());
  }

  /**
   * Senden einer Nachricht inklusive Attachments an den XTA-Broker
   *
   * @param reader          Identität des Leser
   * @param deliveryOptions Sendeoptionen
   * @param message         Nachricht
   * @param attachments     Anhänge
   * @return
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   * @throws MessageSchemaViolationException
   * @throws ParameterIsNotValidException
   * @throws MessageVirusDetectionException
   * @throws SyncAsyncException
   * @throws InvalidMessageIDException
   * @throws ClientException
   */
  public MessageTransportReport sendMessage(final Identifier reader, final DeliveryOptions deliveryOptions, final Message message,
      final List<Attachment> attachments)
      throws ClientException, XTAWSTechnicalProblemException, PermissionDeniedException, ParameterIsNotValidException, MessageSchemaViolationException, MessageVirusDetectionException, SyncAsyncException, InvalidMessageIDException {

    assertRoleSenderServiceAreInitialized();
    final Stopwatch sw = Stopwatch.createStarted();
    final PartyType author = createPartyType(this.identity);

    // 1. checkAccountActive
    managementServiceClient.checkAccountActive(author);

    // 2. lookupService
    assertServiceIsAvailable(deliveryOptions, reader, author);

    // 3. createMessageId
    final AttributedURIType messageId = managementServiceClient.createMessageId(author);

    // 4. sendMessage
    final MessageMetaData messageMetaData = createMessageMetaData(this.identity, reader, messageId, deliveryOptions, message);

    final GenericContentContainer genericContentContainer = createGenericContentContainer(this, deliveryOptions, message,
        attachments);

    //  - Die Funktionalität X509TokenContainer wurde hier nicht umgesetzt. Daher Übergabe von null.
    sendServiceClient.sendMessage(genericContentContainer, messageMetaData, null);

    // 5. getMessageTransportReport
    final MessageTransportReport report;
    if (deliveryOptions.isCheckDeliveryStatus()) {
      final ExecutorService executorService = Executors.newFixedThreadPool(1);
      final Future<MessageTransportReport> messageTransportReportFuture = executorService
          .submit(new CheckMessageDelivery4Status(messageId.getValue(), deliveryOptions));
      executorService.shutdown();
      try {
        report = messageTransportReportFuture.get(deliveryOptions.getCheckStatusMaxWaitTimeSecconds(), TimeUnit.SECONDS);
      } catch (InterruptedException | ExecutionException | TimeoutException e) {
        throw new ClientException(
            "Nachricht: " + messageId.getValue() + " hat nicht den DeliveryStatus " + deliveryOptions.getCheckStatus().name(), e);
      }
    } else {
      report = new MessageTransportReport(messageId.getValue(), managementServiceClient.getTransportReport(messageId, author));
    }

    log.info("sendMessage: {}ms", sw.elapsed(TimeUnit.MILLISECONDS));
    return report;
  }

  /**
   * Zurückziehen einer Nachricht
   *
   * @param messageId NachrichtenID
   * @throws CancelDeniedException
   * @throws ParameterIsNotValidException
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   * @throws InvalidMessageIDException
   */
  public void cancelMessage(final String messageId)
      throws PermissionDeniedException, XTAWSTechnicalProblemException, InvalidMessageIDException, CancelDeniedException, ParameterIsNotValidException {
    final Stopwatch sw = Stopwatch.createStarted();
    final PartyType author = createPartyType(this.identity);

    //1. checkAccountActive
    managementServiceClient.checkAccountActive(author);

    //2. getMessageTransportReport
    final TransportReport report = managementServiceClient.getTransportReport(createAttributedURIType(messageId), author);

    //3. cancelMessage
    managementServiceClient.cancelMessage(createAttributedURIType(messageId), author);
    log.info("cancelMessage: {}ms", sw.elapsed(TimeUnit.MILLISECONDS));
  }

  /**
   * Abholen von Transport-Reports eines Postfachs
   *
   * @param filter Nachrichtenfilter
   * @return
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   * @throws InvalidMessageIDException
   */

  public List<MessageTransportReport> getMessageTransportReportList(final MessageFilter filter)
      throws ClientException, PermissionDeniedException, XTAWSTechnicalProblemException, InvalidMessageIDException {
    assertRoleReceiverServiceAreInitialized();
    final Stopwatch sw = Stopwatch.createStarted();
    final List<MessageTransportReport> messageTransportReportList = new ArrayList<>();
    final Holder<MsgBoxResponseType> fetchResponseHeader = new Holder<>();
    final MsgStatusListType listType = msgBoxServiceClient
        .getStatusList(createMsgBoxStatusListRequestType(filter), createPartyType(this.identity), fetchResponseHeader);

    for (final MessageMetaData messageMetaData : listType.getMessageMetaData()) {
      final AttributedURIType messageId = messageMetaData.getMsgIdentification().getMessageID();

      final TransportReport transportReport = managementServiceClient
          .getTransportReport(messageId, createPartyType(this.identity));
      messageTransportReportList.add(new MessageTransportReport(messageId.getValue(), transportReport));
    }

    log.info("getMessageTransportReportList: {}ms", sw.elapsed(TimeUnit.MILLISECONDS));
    return messageTransportReportList;
  }

  /**
   * holen eines Transport-Reports zu einer Nachricht
   *
   * @param messageId NachrichtenID
   * @return
   * @throws InvalidMessageIDException
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   */
  public MessageTransportReport getMessageTransportReport(final String messageId)
      throws InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException {
    final Stopwatch sw = Stopwatch.createStarted();
    final AttributedURIType attributedURIType = createAttributedURIType(messageId);
    final PartyType author = createPartyType(this.identity);
    final MessageTransportReport report = new MessageTransportReport(messageId,
        managementServiceClient.getTransportReport(attributedURIType, author));
    log.info("getMessageTransportReport: {}ms", sw.elapsed(TimeUnit.MILLISECONDS));
    return report;
  }

  /**
   * Abholen einer Nachricht
   *
   * @param messageId NachrichtenID
   * @return
   * @throws InvalidMessageIDException
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   * @throws ClientException
   */
  public GetMessageResult getMessage(final String messageId)
      throws InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException, ClientException {
    assertRoleReceiverServiceAreInitialized();
    final Stopwatch sw = Stopwatch.createStarted();
    final Holder<MessageMetaData> messageMetaData = new Holder<>();
    final Holder<MsgBoxResponseType> fetchResponseHeader = new Holder<>();

    final GenericContentContainer genericContentContainer = msgBoxServiceClient
        .getMessage(createMsgBoxFetchRequest(messageId), createPartyType(this.identity), messageMetaData, fetchResponseHeader);

    final MessageTransportReport messageTransportReport = this.getMessageTransportReport(messageId);

    final String msgBoxRequestID;

    //  Das Element <xs:attribute name="MsgBoxRequestID" type="xs:anyURI" use="required"/> wird nicht bei allen Nachrichtenvermittlern geliefert.
    //  Folgender Coide wird temporär benötigt um ein valides Response zu erhalten.
    //  - <xs:attribute name="MsgBoxRequestID" type="xs:anyURI" use="required"/>
    //  Mit "not-implement-by-server" kann natürlich aber auch nicht quittiert werden.
    // Dieser Code kann entfernt werden sobald entsprechender Broker valides Response erzeugt.
    if (fetchResponseHeader.value == null) {
      msgBoxRequestID = "not-implement-by-server"; //messageId;
    } else {
      msgBoxRequestID = fetchResponseHeader.value.getMsgBoxRequestID();
    }

    final String serviceProfileName = (messageMetaData.value.getDeliveryAttributes() != null ? messageMetaData.value
        .getDeliveryAttributes().getServiceQuality() : null);
    final ContentContainer contentContainer = getContentContainer(this, genericContentContainer, serviceProfileName);
    final GetMessageResult result = new GetMessageResult(getMessage(contentContainer),
        getAttachments(contentContainer), messageTransportReport, msgBoxRequestID);

    log.info("getMessage: {} ms", sw.elapsed(TimeUnit.MILLISECONDS));

    return result;
  }

  /**
   * Abholen einer Nachricht und gleichzeitiges bEstätigen des Empfangs
   *
   * @param messageId NachrichtenID
   * @return
   * @throws InvalidMessageIDException
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   * @throws ClientException
   */
  public GetMessageResult getMessageAndClose(final String messageId)
      throws InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException, ClientException {
    final GetMessageResult result = getMessage(messageId);
    close(result.getMessageTransportReport().getMessageId(), result.getRequestHandle());
    final MessageTransportReport report = getMessageTransportReport(result.getMessageTransportReport().getMessageId());
    try {
      FieldUtils.writeField(result, "messageTransportReport", report, true);
    } catch (final IllegalAccessException e) {
      throw new ClientException(e.getMessage(), e);
    }
    return result;
  }


  /**
   * Bestätigung des Empfang einer Nachricht
   *
   * @param messageId     NachrichtenID
   * @param requestHandle
   * @throws InvalidMessageIDException
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   */
  public void close(final String messageId, final String requestHandle)
      throws InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException {
    assertRoleReceiverServiceAreInitialized();
    final Stopwatch sw = Stopwatch.createStarted();
    msgBoxServiceClient.close(createMsgBoxCloseRequestType(messageId, requestHandle), createPartyType(this.identity));
    log.info("close: {}ms", sw.elapsed(TimeUnit.MILLISECONDS));
  }

  /**
   * Bestätigung des Empfang einer Nachricht mit nachrichtenId
   *
   * @param messageId NachrichtenID
   * @throws InvalidMessageIDException
   * @throws PermissionDeniedException
   * @throws XTAWSTechnicalProblemException
   */
  public void close(final String messageId)
      throws InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException {
    close(messageId, messageId);
  }


  EncryptionService getEncryptionService(final String serviceProfilename) throws ClientException {
    for (final EncryptionService service : encryptionServices) {
      if (service.getServiceProfilname().equals(serviceProfilename)) {
        return service;
      }
    }
    throw new ClientException("Encryption-Service für " + serviceProfilename + " nicht registriert.");
  }


  @RequiredArgsConstructor
  private class CheckMessageDelivery4Status implements Callable<MessageTransportReport> {

    private final String messageId;
    private final DeliveryOptions deliveryOptions;

    @Override
    public MessageTransportReport call()
        throws ClientException, InterruptedException, InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException {

      for (int i = 0; i < deliveryOptions.getCheckStatusAttempts(); i++) {
        final MessageTransportReport messageTransportReport = new MessageTransportReport(messageId,
            managementServiceClient.getTransportReport(createAttributedURIType(messageId), createPartyType(identity)));
        if (messageTransportReport.getDeliveryStatus().ordinal() >= deliveryOptions.getCheckStatus().ordinal()) {
          return messageTransportReport;
        }

        Thread.sleep(deliveryOptions.getCheckStatusWaitTimeSecconds() * 1000);
      }

      throw new ClientException(
          "Nachricht: " + messageId + " hat nicht den DeliveryStatus " + deliveryOptions.getCheckStatus().name());
    }
  }

  private void assertRoleSenderServiceAreInitialized() {
    if (sendServiceClient == null) {
      throw new IllegalStateException("sendServiceClient ist nicht initializiert");
    }
  }

  private void assertRoleReceiverServiceAreInitialized() {
    if (msgBoxServiceClient == null) {
      throw new IllegalStateException("msgBoxServiceClient ist nicht initializiert");
    }
  }

  private void assertServiceIsAvailable(DeliveryOptions deliveryOptions, Identifier reader, PartyType author)
      throws ClientException, ParameterIsNotValidException, PermissionDeniedException, XTAWSTechnicalProblemException {

    final LookupServiceResponse lookupServiceResponse = managementServiceClient
        .lookupService(createLookupServiceRequest(deliveryOptions, reader), author);

    if (lookupServiceResponse != null && lookupServiceResponse.getLookupServiceResultList() != null
        && lookupServiceResponse.getLookupServiceResultList().getLookupServiceResult() != null) {
      for (LookupServiceResultType res : lookupServiceResponse.getLookupServiceResultList().getLookupServiceResult()) {
        if (!res.getIsServiceAvailableValue().isServiceIsAvailable()) {
          throw new ClientException("minimum one requested ServiceType [" + res.getServiceType() + "] is not available");
        }
      }

    } else {
      throw new ClientException("broker sent null-response, expected is a valid LookupServiceResponse");
    }

  }
}
