/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client;

import de.xoev.xta.client.config.HttpClientPolicy;
import de.xoev.xta.client.config.HttpProxy;
import de.xoev.xta.client.config.Keystore;
import de.xoev.xta.client.config.XtaClientSettings;
import de.xoev.xta.client.encyrption.EncryptionService;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.exception.ClientInitializationException;
import de.xoev.xta.client.model.Identifier;
import java.net.MalformedURLException;
import java.net.URL;

import lombok.AccessLevel;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.Setter;

/**
 * Config-Holder for the Xta-Client.
 */
@Getter
@RequiredArgsConstructor
@Setter(value = AccessLevel.PROTECTED)
public class XtaClientConfig {

  private Identifier author;

  private URL managementServiceUrl;
  private URL sendServiceUrl;
  private URL msgBoxServiceUrl;
  private HttpClientPolicy httpClientPolicy;
  private HttpProxy httpProxy;
  private Keystore clientCertKeystore;
  private Keystore trustCertKeystore;

  private boolean validation;
  private boolean logSoapRequest;
  private boolean logSoapResponse;

  private EncryptionService[] encryptionServices;

  public static class XtaClientConfigBuilder {

    private XtaClientConfig config = new XtaClientConfig();

    public XtaClientConfigBuilder withValidation() {
      config.setValidation(true);
      return this;
    }

    public XtaClientConfigBuilder withSoapLogging() {
      config.setLogSoapRequest(true);
      config.setLogSoapResponse(true);
      return this;
    }

    public XtaClientConfigBuilder withHttpClientPolicy(final HttpClientPolicy policy) {
      config.setHttpClientPolicy(policy);
      return this;
    }

    public XtaClientConfigBuilder withHttpProxy(final HttpProxy proxy) {
      config.setHttpProxy(proxy);
      return this;
    }

    public XtaClientConfigBuilder withEncryptionServices(final EncryptionService... encryptionServices) {
      config.setEncryptionServices(encryptionServices);
      return this;
    }

    public XtaClientConfig build() {
      return this.config;
    }

  }

  public static XtaClientConfigBuilder create(final URL manangementServiceUrl, final URL sendServiceUrl,
      final URL msgBoxServiceUrl, final Identifier author, final String clientCertKeystore,
      final String clientCertKeyStorePassword, final String trustCertKeystore, final String trustCertKeyStorePassword) {

    final XtaClientConfigBuilder builder = new XtaClientConfigBuilder();

    builder.config.setMsgBoxServiceUrl(msgBoxServiceUrl);
    builder.config.setManagementServiceUrl(manangementServiceUrl);
    builder.config.setSendServiceUrl(sendServiceUrl);
    builder.config.setAuthor(author);
    builder.config
        .setClientCertKeystore(Keystore.builder().path(clientCertKeystore).password(clientCertKeyStorePassword).build());
    builder.config.setTrustCertKeystore(Keystore.builder().path(trustCertKeystore).password(trustCertKeyStorePassword).build());

    return builder;
  }

  public static XtaClientConfigBuilder create(final XtaClientSettings settings) throws ClientException {
    return create(Identifier.from(settings), settings);
  }

  public static XtaClientConfigBuilder create(final Identifier author, final XtaClientSettings settings) throws ClientException {
    try {
      // Endpoint-URLs
      final URL manangementServiceUrl = new URL(settings.get(XtaClientSettings.BROKER_MANAGEMENT_SERVICE_ENDPONIT));
      final URL sendServiceUrl = new URL(settings.get(XtaClientSettings.BROKER_SEND_SERVICE_ENDPOINT));
      final URL msgBoxServiceUrl = new URL(settings.get(XtaClientSettings.BROKER_MSGBOX_SERVICE_ENDPOINT));

      // Clientcert
      final String clientCertKeystore = settings.get(XtaClientSettings.CLIENT_KEYSTORE_CERT_PATH);
      final String clientCertKeyStorePassword = settings.get(XtaClientSettings.CLIENT_KEYSTORE_CERT_PASSPHRASE);

      // Servercert
      final String trustCertKeystore = settings.get(XtaClientSettings.BROKER_KEYSTORE_TRUSTCERT_PATH);
      final String trustCertKeyStorePassword = settings.get(XtaClientSettings.BROKER_KEYSTORE_TRUSTCERT_PASSPHRASE);

      return create(manangementServiceUrl, sendServiceUrl, msgBoxServiceUrl, author, clientCertKeystore,
          clientCertKeyStorePassword, trustCertKeystore, trustCertKeyStorePassword);

    } catch (MalformedURLException e) {
      throw new ClientInitializationException("Some URLs could not be parsed: " + e.getMessage(), e);
    }
  }
}
