/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.core;

import com.google.common.base.Joiner;
import com.google.common.collect.Lists;
import de.xoev.xta.client.config.HttpClientPolicy;
import de.xoev.xta.client.config.HttpProxy;
import de.xoev.xta.client.config.Keystore;
import de.xoev.xta.client.exception.ClientInitializationException;
import java.net.URL;
import java.util.List;

import lombok.AccessLevel;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;

/**
 * Konstruiert eine Instanz des {@link WsClient}.
 */
@Getter(AccessLevel.PACKAGE)
@Slf4j
public abstract class ServiceBuilder<T extends WsClient> {

  @Getter(AccessLevel.NONE)
  final WsClientConfig clientConfiguration;

  /**
   * Initialisiert einen {@link ServiceBuilder} mit ServiceUrl-Host und ServiceUrl-Port.
   *
   * @param serviceUrl URL des Endpoints
   */
  ServiceBuilder(final URL serviceUrl) {
    this.clientConfiguration = new WsClientConfig();
    this.clientConfiguration.setServiceUrl(serviceUrl);
  }

  /**
   * Sezt Trust-Keystore des {@link WsClient}s.
   *
   * @param keystore Trust-Keystore
   * @return {@link ServiceBuilder}
   */
  public ServiceBuilder<T> setTrustCertKeystore(final Keystore keystore) {
    this.clientConfiguration.setTrustCertKeystore(keystore);
    return this;
  }

  /**
   * Setzt den Client-Keystore des {@link WsClient}s.
   *
   * @param keystore Client-Keystore
   * @return {@link ServiceBuilder}
   */
  public ServiceBuilder<T> setClientCertKeystore(final Keystore keystore) {
    this.clientConfiguration.setClientCertKeystore(keystore);
    return this;
  }

  /**
   * Setzt den Proxy Host.
   *
   * @param proxy Proxy Host
   * @return {@link ServiceBuilder}
   */
  public ServiceBuilder<T> setHttpProxy(final HttpProxy proxy) {
    this.clientConfiguration.setHttpProxy(proxy);
    return this;
  }

  /**
   * Setzt XSD Validation
   *
   * @return {@link ServiceBuilder}
   */
  public ServiceBuilder<T> setValidation(final boolean validation) {
    this.clientConfiguration.setValidation(validation);
    return this;
  }

  public ServiceBuilder<T> enableValidation() {
    setValidation(true);
    return this;
  }

  public ServiceBuilder<T> enableSoapLogging() {
    setSoapRequestLogging(true);
    setSoapResponseLogging(true);
    return this;
  }

  /**
   * Setzt SoapRequest Logging
   *
   * @return {@link ServiceBuilder}
   */
  public ServiceBuilder<T> setSoapRequestLogging(final boolean logging) {
    this.clientConfiguration.setLogSoapRequest(logging);
    return this;
  }

  /**
   * Setzt SoapResponse Logging
   *
   * @return {@link ServiceBuilder}
   */
  public ServiceBuilder<T> setSoapResponseLogging(final boolean logging) {
    this.clientConfiguration.setLogSoapResponse(logging);
    return this;
  }

  /**
   * Setzt eine HttpClient-Policy
   *
   * @param policy HttpClientPolicy
   * @return {@link ServiceBuilder}
   */

  public ServiceBuilder<T> setHttpClientPolicy(final HttpClientPolicy policy) {
    this.clientConfiguration.setHttpClientPolicy(policy);
    return this;
  }

  void validateConfiguration() {
    final List<String> errorMsgs = Lists.newArrayList();
    ServiceBuilder.checkNotNull("Service-URL", this.clientConfiguration.getServiceUrl(), errorMsgs);

    // Entweder beide Angaben leer oder beide gesetzt
    if (this.clientConfiguration.getHttpProxy() != null) {
      if (this.clientConfiguration.getHttpProxy().getHost() == null ^ this.clientConfiguration.getHttpProxy().getPort() == 0) {
        errorMsgs.add("Angaben zum Proxy (Host, Port) müssen vollständig gesetzt sein");
      }
    }

    this.checkKeystoreFile("Client-Keystore", this.clientConfiguration.getClientCertKeystore(), errorMsgs);
    this.checkKeystoreFile("Trust-Keystore", this.clientConfiguration.getTrustCertKeystore(), errorMsgs);

    if (errorMsgs != null && !errorMsgs.isEmpty()) {
      throw new IllegalArgumentException(Joiner.on("\n").join(errorMsgs));
    }
  }

  private static void checkNotNull(final String identifier, final Object value, final List<String> errorMsgs) {
    if (value == null) {
      errorMsgs.add(identifier + ": Wert muss gesetzt sein");
    }
  }

  private void checkKeystoreFile(final String keyStoreIdentifier, final Keystore keyStore,
      final List<String> errorMsgs) {
    if (StringUtils.isEmpty(keyStore.getPath())) {
      errorMsgs.add(keyStoreIdentifier + ": Pfad muss gesetzt sein");
    }
    ServiceBuilder.checkNotNull(keyStoreIdentifier + " Passwort", keyStore.getPassword(), errorMsgs);
  }

  public abstract T build() throws ClientInitializationException;
}
