/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.core;

import de.xoev.xta.client.config.HttpClientPolicy;
import de.xoev.xta.client.config.HttpProxy;
import de.xoev.xta.client.exception.ClientInitializationException;
import de.xoev.xta.client.util.KeyStoreUtil;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509TrustManager;
import javax.xml.ws.BindingProvider;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.cxf.common.util.StringUtils;
import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.ext.logging.LoggingInInterceptor;
import org.apache.cxf.ext.logging.LoggingOutInterceptor;
import org.apache.cxf.frontend.ClientProxy;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;

/**
 * Basisklasse für gewrapte Serviceklassen
 */
@Slf4j
abstract class WsClient<T> {

  WsClientConfig clientConfiguration;

  T port;

  abstract void initializePort(final URL url);

  void initializeService() throws ClientInitializationException {

    final String logMessage =
        "JarInfo: " + this.getClass().getName() + ",\nConfiguration: " + this.clientConfiguration + "," + "\ninitializeService";

    try {
      final String serviceUrl = this.clientConfiguration.getServiceUrl().toString();
      WsClient.log.debug("Initialize Service {}", serviceUrl);
      final URL url = this.getClass().getClassLoader().getResource("wsdl/XTA.wsdl");
      WsClient.log.debug("url: {}", url);
      this.initializeServiceEndpoint(url);
      this.initializeTransportSecurity();
      this.initializeHttpClientPolicy();
      this.initializeProxy();
      WsClient.log.info("{} sucessfully.", logMessage);
    } catch (final Exception ex) {
      WsClient.log.error("{} failed. Reason: {}", logMessage, ex.getMessage());
      throw new ClientInitializationException(ex.getMessage(), ex);
    }
  }

  private void initializeServiceEndpoint(final URL url) {
    initializePort(url);
    final BindingProvider bp = (BindingProvider) this.port;

    bp.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, this.clientConfiguration.getServiceUrl().toString());

    WsClient.log.debug("Set EnpointUrl to:  {}", this.clientConfiguration.getServiceUrl().toString());

    // Empty xmlns="" entfernen -> XSD-Modellierung Annonymous Inner Type
    // ClientProxy.getClient(port).getOutInterceptors().add(new EmptyNamespaceMessageOutInterceptor());

    initializeValidation();

    if (this.clientConfiguration.isLogSoapRequest()) {
      LoggingOutInterceptor loi = new LoggingOutInterceptor();
      loi.setPrettyLogging(true);
      ClientProxy.getClient(port).getOutInterceptors().add(loi);
      WsClient.log.debug("Set SoapReqeust-Logging: {}", clientConfiguration.isLogSoapRequest());
    }

    if (this.clientConfiguration.isLogSoapResponse()) {
      LoggingInInterceptor lii = new LoggingInInterceptor();
      lii.setPrettyLogging(true);
      ClientProxy.getClient(port).getInInterceptors().add(lii);
      WsClient.log.debug("Set SoapResponse-Logging: {}", clientConfiguration.isLogSoapResponse());
    }
  }

  private void initializeHttpClientPolicy() {
    final HttpClientPolicy policy = this.clientConfiguration.getHttpClientPolicy();
    if (policy != null) {
      final Client client = ClientProxy.getClient(this.port);
      final HTTPConduit httpConduit = (HTTPConduit) client.getConduit();
      final HTTPClientPolicy httpClientPolicy = new HTTPClientPolicy();
      if (policy.getConnectionTimeout() != null) {
        httpClientPolicy.setConnectionTimeout(policy.getConnectionTimeout());
      }
      if (policy.getAllowChunking() != null) {
        httpClientPolicy.setAllowChunking(policy.getAllowChunking());
      }
      if (policy.getReceiveTimeOut() != null) {
        httpClientPolicy.setReceiveTimeout(policy.getReceiveTimeOut());
      }
      httpConduit.setClient(httpClientPolicy);
      WsClient.log.debug("Intitialize HttpClientPolicy");
    }
  }

  private void initializeTransportSecurity() throws ClientInitializationException {
    if (this.clientConfiguration.getClientCertKeystore() != null) {
      final Client client = ClientProxy.getClient(this.port);
      final HTTPConduit httpConduit = (HTTPConduit) client.getConduit();
      final TLSClientParameters parameters = new TLSClientParameters();
      parameters.setSSLSocketFactory(this.createSSLContext().getSocketFactory());
      httpConduit.setTlsClientParameters(parameters);
      WsClient.log.debug("Intitialize TransportSecurity");
    }
  }

  private void initializeProxy() {
    final HttpProxy proxy = this.clientConfiguration.getHttpProxy();
    if (proxy != null) {
      final Client client = ClientProxy.getClient(this.port);
      final HTTPConduit httpConduit = (HTTPConduit) client.getConduit();
      if (!StringUtils.isEmpty(proxy.getHost())) {
        httpConduit.getClient().setProxyServer(proxy.getHost());
      }
      if (proxy.getPort() != null) {
        httpConduit.getClient().setProxyServerPort(proxy.getPort());
      }
      if (!StringUtils.isEmpty(proxy.getUsername())) {
        httpConduit.getProxyAuthorization().setUserName(proxy.getUsername());
      }
      if (!StringUtils.isEmpty(proxy.getPassword())) {
        httpConduit.getProxyAuthorization().setPassword(proxy.getPassword());
      }
      WsClient.log.debug("Intitialize Proxy");
    }
  }

  private SSLContext createSSLContext() throws ClientInitializationException {
    try {

      // Server
      final KeyStore trustStore = KeyStoreUtil
          .load(this.clientConfiguration.getTrustCertKeystore().getPath(),
              this.clientConfiguration.getTrustCertKeystore().getPassword());

      // client
      final KeyStore keyStore = KeyStoreUtil
          .load(this.clientConfiguration.getClientCertKeystore().getPath(),
              this.clientConfiguration.getClientCertKeystore().getPassword());

      //Client-Lib Truststore
      final TrustManagerFactory tmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
      tmf.init(trustStore);

      //Default JDK Truststore
      final TrustManagerFactory defaultTmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
      defaultTmf.init((KeyStore) null);

      //Client-Lib-Keystore
      final KeyManagerFactory kmf = KeyManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
      kmf.init(keyStore, this.clientConfiguration.getClientCertKeystore().getPassword().toCharArray());

      final SSLContext sslContext = SSLContext.getInstance("TLSv1.2");

      CustomTruststoreManager customTruststoreManager = new CustomTruststoreManager(getX509TrustMananger(defaultTmf),
          getX509TrustMananger(tmf));

      sslContext.init(kmf.getKeyManagers(), new TrustManager[]{customTruststoreManager}, new SecureRandom());

      return sslContext;
    } catch (UnrecoverableKeyException | NoSuchAlgorithmException | KeyStoreException | KeyManagementException e) {
      throw new ClientInitializationException("Exception in createSSLContext: " + e.getMessage(), e);
    }
  }

  private X509TrustManager getX509TrustMananger(TrustManagerFactory tmf) {
    for (TrustManager tm : tmf.getTrustManagers()) {
      if (tm instanceof X509TrustManager) {
        return (X509TrustManager) tm;
      }
    }

    return null;
  }

  public void enableValidation() {
    clientConfiguration.setValidation(true);
    initializeValidation();
  }

  public void disableValidation() {
    clientConfiguration.setValidation(false);
    initializeValidation();
  }

  private void initializeValidation() {
    final BindingProvider bp = (BindingProvider) this.port;

    if (this.clientConfiguration.isValidation()) {
      bp.getRequestContext().put("schema-validation-enabled", true);
    } else {
      bp.getRequestContext().put("schema-validation-enabled", false);
    }

    WsClient.log.debug("Set Enpoint-Validation: {}", clientConfiguration.isValidation());
  }

  @RequiredArgsConstructor
  private class CustomTruststoreManager implements X509TrustManager {

    final X509TrustManager defaultTm;
    final X509TrustManager myTm;

    @Override
    public X509Certificate[] getAcceptedIssuers() {
      // If you're planning to use client-cert auth,
      // merge results from "defaultTm" and "myTm".
      return defaultTm.getAcceptedIssuers();
    }

    @Override
    public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
      try {
        myTm.checkServerTrusted(chain, authType);
      } catch (CertificateException e) {
        defaultTm.checkServerTrusted(chain, authType);
      }
    }

    @Override
    public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {
      // If you're planning to use client-cert auth,
      // do the same as checking the server.
      defaultTm.checkClientTrusted(chain, authType);
    }
  }
}
