/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.model;

import de.xoev.xta.client.codes.BusinessScenario;
import java.util.Date;

import lombok.AccessLevel;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import org.apache.commons.lang3.NotImplementedException;

/**
 * Model-Klasse, repräsentiert die Optionen eines Transportauftrages. Wird vor dem Nachrichtenversand erzeugt und beim Senden
 * mitgegeben.
 */
@Getter
@NoArgsConstructor(access = AccessLevel.PUBLIC)
public final class DeliveryOptions {

  private Date origin;
  private Date sendNotBefore;
  private Date obsoleteAfter;
  private DeliveryStatus checkStatus;
  private int checkStatusAttempts = 5;
  private int checkStatusWaitTimeSecconds = 10;
  private int checkStatusMaxWaitTimeSecconds = 60;
  private boolean sendSync = false;
  private String service;
  private Boolean isTestMessage = false;
  private String businessScenario;
  private Identifier replyTo;

  @Setter(value = AccessLevel.NONE)
  private boolean definedBusinessScenario;


  private String encryptionServiceProfilename;

  /**
   * Konstruktur
   *
   * @param service Name/Bezeichnung des service (als URI)
   * @param bs      Name/Bezeichnung des Businessszenario
   */
  public DeliveryOptions(final String service, final BusinessScenario bs) {
    this.origin = new Date();
    this.service = service;
    withBusinessScenario(bs);
  }

  /**
   * Konstruktur
   *
   * @param service Name/Bezeichnung des service (als URI)
   * @param bs      Name/Bezeichnung des Businessszenario
   */
  public DeliveryOptions(final String service, final String bs) {
    this.origin = new Date();
    this.service = service;
    withBusinessScenario(bs);
  }

  public DeliveryOptions setSendNotBefore(final Date date) {
    this.sendNotBefore = date;
    return this;
  }

  public DeliveryOptions setObsoleteAfter(final Date date) {
    this.obsoleteAfter = date;
    return this;
  }

  public DeliveryOptions withCheckDeliveryStatus(final DeliveryStatus status) {
    this.checkStatus = status;
    return this;
  }

  public DeliveryOptions withCheckDeliveryStatus(final DeliveryStatus status, final int attempts, final int waitTime,
      final int maxWaitTime) {
    this.checkStatus = status;
    this.checkStatusAttempts = attempts;
    this.checkStatusWaitTimeSecconds = waitTime;
    this.checkStatusMaxWaitTimeSecconds = maxWaitTime;
    return this;
  }

  public DeliveryOptions withSendSync() {
    throw new NotImplementedException("currently not implemented");
  }

  public DeliveryOptions withReplyTo(final Identifier identifier) {
    this.replyTo = identifier;
    return this;
  }


  public boolean isCheckDeliveryStatus() {
    return checkStatus != null;
  }

  /**
   * Kennzeichnen des Transportauftrages als Test.
   *
   * @return this
   */
  public DeliveryOptions asTestMessage() {
    isTestMessage = true;
    return this;
  }

  /**
   * Angabe des Fachverfahrens am Auftrag.
   *
   * @param bs bezeichnung des Fachverfahrens {@link BusinessScenario}
   * @return this
   */
  public DeliveryOptions withBusinessScenario(final BusinessScenario bs) {
    businessScenario = bs.getCode();
    definedBusinessScenario = true;
    return this;
  }

  /**
   * Angabe des Fachkontextes/Verfahrens am Transportauftrag.
   *
   * @param bs Bezeichnung des Fachverfahrens als String. Kann verwendet werden wenn die Enum das gewünschte Szenario noch nicht
   *           enthält
   * @return this
   */
  public DeliveryOptions withBusinessScenario(final String bs) {
    businessScenario = bs;
    definedBusinessScenario = false;
    return this;
  }


  public DeliveryOptions withEncryptionServiceProfile(final String profilename) {
    encryptionServiceProfilename = profilename;
    return this;
  }

}
