/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.model;

import de.xoev.xta.client.codes.IdentifierType;
import de.xoev.xta.client.config.XtaClientSettings;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.exception.ClientInitializationException;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.StringUtils;

/**
 * Modelklasse, repräsentiert einen Kommunikationspartner, bestehend aus Kennung, Typ, Kategorie (optinal) und einen
 * benutzerfreundlichen Namen (optional). Diese Klasse definiert nicht die (technische) Rolle des Partners wie Leser oder Autor,
 * bzw. Sender oder Empfänger, sondern kann für beides stehen.
 */
@Getter
@NoArgsConstructor(access = AccessLevel.PUBLIC)
public class Identifier {

  // Pflichtfeld: die Kennung, üblich <Prefix>:<Kennung>
  private String id;

  // Pflichtfeld: Typ der Kennung, z.B.: XOEV oder JUSTIZ
  private String type;

  // Optional: benutzerfreundlicher Name, "Anzeigename"
  private String name;

  // Optional: Konkretesiert die fachliche Rolle eines Partners genauer (e.g. "buyer", "Meldehörde", "Standesamt"...)
  private String category;

  /**
   * Erzeugen eines Identifier mit seinen minimalen Pflichtfeldern.
   *
   * @param id   die ("technische") Kennung des Kommunikationspartners, zb. xdo:11009991
   * @param type Typ des Kommunikationspartners, zb. XOEV
   */
  public Identifier(final String id, String type) throws ClientInitializationException {
    this.id = id;
    this.type = type;

    assertIsValid(this, "Constructor");
  }

  /**
   * Erzeugen eines Identifier mit seinen minimalen Pflichtfeldern.
   *
   * @param id   die ("technische") Kennung des Kommunikationspartners, zb. xdo:11009991
   * @param type Typ des Kommunikationspartners, zb. XOEV, {@link IdentifierType}
   */
  public Identifier(final String id, IdentifierType type) throws ClientInitializationException {
    this(id, type.getCode());
  }

  /**
   * Statische Fabrik-Methode zum Erzeugen einer Identifier-Instanz aus den Einstellungen (settings.properties). Typischerweise
   * ist dies die Identität des XtaClients selbst.
   *
   * @return eine Identifier-Instanz gefüllt mit den Daten der Einstellungen.
   * @throws ClientException wenn die erzeugte Instans nicht valide ist oder es probleme mit dem Settingfile gibt (IO)
   */
  public static Identifier from(final XtaClientSettings settings) throws ClientException {
    Identifier target = new Identifier();
    target.id = settings.get(XtaClientSettings.CLIENT_IDENTIFIER_ID_KEY);
    target.type = settings.get(XtaClientSettings.CLIENT_IDENTIFIER_TYPE_KEY);
    target.category = settings.get(XtaClientSettings.CLIENT_IDENTIFIER_CATEGORY_KEY);
    target.name = settings.get(XtaClientSettings.CLIENT_IDENTIFIER_NAME_KEY);

    assertIsValid(target, "Settings");
    return target;
  }

  /**
   * Mit einem optionalen "Nutzerfreundlichen", zb. "Heiko Materny"
   *
   * @param name der Name
   * @return Identifier-Instanz ("sich selbst"") gefüllt mit dem Namen.
   */
  public Identifier withUserFriendlyName(String name) {
    this.name = name;
    return this;
  }

  /**
   * Mit einer optionalen Kategorie, zb. "Standesamt"
   *
   * @param category die Kategorie, e.g. Standesamt
   * @return Identifier-Instanz ("sich selbst"") gefüllt mit der Kategorie.
   */
  public Identifier withCategory(String category) {
    this.name = category;
    return this;
  }

  private static void assertIsValid(Identifier toCheck, String errMsg) throws ClientInitializationException {
    if (StringUtils.isEmpty(toCheck.id)) {
      throw new ClientInitializationException("construction from " + errMsg + " failed, reason: id is empty");
    }

    if (StringUtils.isEmpty(toCheck.type)) {
      throw new ClientInitializationException("construction from " + errMsg + " failed, reason: type is empty");
    }
  }

  /**
   * Enum zur Auswahl der Kommunikationsrolle (spielt nur eine Rolle wenn der innere Nachrichtenkopf verarbeitet wird.)
   */
  public enum CommunicationRole {
    AUTHOR, READER, REPLYTO;
  }
}
