/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.model;

import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.util.XmlUtil;
import java.io.IOException;
import java.math.BigInteger;

import lombok.Getter;
import org.apache.commons.io.IOUtils;
import org.w3c.dom.Document;

/**
 * Model-Klasse, representiert die zu sendene Nachricht.
 */
@Getter
public class Message extends Content {

  private String messageTypeCode;
  private String messageTypePayloadSchema;

  /**
   * Erzeugt eine Nachricht aus Daten und Inhaltstyp.
   *
   * @param contentType Inhaltstyp
   * @param data        Inhalt als Byte-Array
   */
  public Message(final String contentType, final byte[] data) {
    super(contentType, data);
    extractMessageType();
    setSize(BigInteger.valueOf(Long.valueOf(data.length)));
  }

  public Message(final String contentType, final byte[] data, final String contentDescription, final String encoding,
      final String filename, final String id, final String lang, final BigInteger size) {

    this(contentType, data);

    setContentDescription(contentDescription);
    setEncoding(encoding);
    setFilename(filename);
    setId(id);
    setLanguage(lang);
    setSize(size);
    extractMessageType();
  }

  private void extractMessageType() {
    try {
      final Document doc = XmlUtil.toNsAwareDocument(data);
      messageTypeCode = doc.getDocumentElement().getLocalName();
      messageTypePayloadSchema = doc.getDocumentElement().getNamespaceURI();
    } catch (final ClientException e) {
      throw new IllegalArgumentException("messageType und/oder messageTypePayloadSchema konnten nicht ermittelt werden.");
    }
  }

  public Message withContentDescription(final String contentDescription) {
    setContentDescription(contentDescription);
    return this;
  }

  public Message withFilename(final String filename) {
    setFilename(filename);
    return this;
  }

  public Message withEncoding(final String encoding) {
    setEncoding(encoding);
    return this;
  }

  public Message withId(final String id) {
    setId(id);
    return this;
  }

  public Message withLanguage(final String lang) {
    setLanguage(lang);
    return this;
  }

  /**
   * Liefert die gekapselten Daten (byte[]) als String.
   *
   * @return die ganze Nachricht als String
   * @throws ClientException
   */
  public String getDataAsString() throws ClientException {
    try {
      return IOUtils.toString(this.getData(), "UTF-8");
    } catch (IOException e) {
      throw new ClientException("Getting message-data as string failed: " + e.getMessage(), e);
    }
  }
}
