/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.soap;

import de.xoev.xta.client.XtaClient;
import de.xoev.xta.client.XtaClientConfig;
import de.xoev.xta.client.config.XtaClientSettings;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.exception.ClientInitializationException;
import de.xoev.xta.client.util.FileUtil;
import de.xoev.transport.xta._211.CancelDeniedException;
import de.xoev.transport.xta._211.InvalidMessageIDException;
import de.xoev.transport.xta._211.ParameterIsNotValidException;
import de.xoev.transport.xta._211.PermissionDeniedException;
import de.xoev.transport.xta._211.XTAWSTechnicalProblemException;
import java.io.IOException;
import javax.jws.WebService;
import javax.xml.ws.Endpoint;

import lombok.NoArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

/**
 * SOAP-Server implementierung des XTA-Clients
 */
@Slf4j
@NoArgsConstructor
@WebService(serviceName = "xtaclient", endpointInterface = "de.init.xtaclient.soap.XtaClientSoapService")
public class XtaClientSoapServer implements XtaClientSoapService {

  private XtaClient xtaClient;

  public static void main(String args[]) {
    try {
      final CommandLineParser parser = new DefaultParser();
      final CommandLine cmd = parser.parse(prepareOptions(), args, true);
      final XtaClientSoapServer me = new XtaClientSoapServer(XtaClientConfig.create(
          new XtaClientSettings(FileUtil.loadByName(cmd.getOptionValue("settings"))))
          .withSoapLogging()
          .build());
      me.run(Long.valueOf(cmd.getOptionValue("port")));
    } catch (final ParseException | ClientException | IOException e) {
      log.error(e.getMessage(), e);
    }
  }

  public void run(final long port) {
    final String address = "http://localhost:" + String.valueOf(port) + "/xtaclient";
    Endpoint.publish(address, this);
    log.info("SoapServer is running...");
  }

  public XtaClientSoapServer(final XtaClientConfig xtaClientConfig) throws ClientInitializationException {
    this.xtaClient = new XtaClient(xtaClientConfig);
  }

  private static Options prepareOptions() {
    Options options = new Options();
    // Required options
    final Option settingsOption = Option.builder().required()
        .longOpt("settings")
        .desc("path to settings file.")
        .hasArg()
        .build();
    final Option portOption = Option.builder().required()
        .longOpt("port")
        .desc("port.")
        .hasArg()
        .build();
    options.addOption(settingsOption)
        .addOption(portOption);
    return options;
  }

  @Override
  public void close(final String messageId, final String requestHandle)
      throws InvalidMessageIDException, PermissionDeniedException, XTAWSTechnicalProblemException {
    xtaClient.close(messageId, requestHandle);
  }

  @Override
  public void cancelMessage(String messageId)
      throws PermissionDeniedException, XTAWSTechnicalProblemException, InvalidMessageIDException, CancelDeniedException, ParameterIsNotValidException {
    xtaClient.cancelMessage(messageId);
  }


}
