/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.util;

import com.google.common.base.Stopwatch;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.ValidationEvent;
import javax.xml.bind.util.ValidationEventCollector;
import javax.xml.transform.stream.StreamSource;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.extern.slf4j.Slf4j;

/**
 * Utility-Klasse für Jaxb-Marshaling/Unmarshalling
 */
@Slf4j
@SuppressWarnings("checkstyle:HideUtilityClassConstructor")
@NoArgsConstructor(access = AccessLevel.NONE)
public final class JaxbUtil {

  private static final Map<Class<?>, JAXBContext> jaxbContextMap = new HashMap<>();

  private static JAXBContext getJaxbContext(final Class<?> clazz) throws JAXBException {
    if (!jaxbContextMap.containsKey(clazz)) {
      jaxbContextMap.put(clazz, JAXBContext.newInstance(clazz));
    }
    return jaxbContextMap.get(clazz);
  }

  public static <T> String serialize(final T t, final Class<?> clazz) throws JAXBException {
    final JaxbValidationEventCollector validationCollector = new JaxbValidationEventCollector();
    final String ret;
    try {
      final Marshaller jaxbMarshaller = getJaxbContext(clazz).createMarshaller();
      // output pretty printed
      jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
      jaxbMarshaller.setProperty(Marshaller.JAXB_ENCODING, "UTF-8");
      jaxbMarshaller.setEventHandler(validationCollector);
      Stopwatch s = Stopwatch.createStarted();
      try (final StringWriter output = new StringWriter()) {
        jaxbMarshaller.marshal(t, output);
        ret = output.toString();
      }
      log.debug("serialize time: " + s.elapsed(TimeUnit.MILLISECONDS));
    } catch (final IOException e) {
      throw new JAXBException(e.getMessage(), e);
    } finally {
      validationCollector.logEvents("serialize: " + clazz.getName());
    }
    return ret;
  }

  public static <T> T deserialize(final String content, final Class<?> clazz) throws JAXBException {
    final T object;
    final JaxbValidationEventCollector validationCollector = new JaxbValidationEventCollector();
    try {
      final Unmarshaller jaxbUnmarshaller = getJaxbContext(clazz).createUnmarshaller();
      jaxbUnmarshaller.setEventHandler(validationCollector);
      final Stopwatch s = Stopwatch.createStarted();
      try (StringReader r = new StringReader(content)) {
        object = (T) jaxbUnmarshaller.unmarshal(new StreamSource(r));
      }
      log.debug("deserialize time: " + s.elapsed(TimeUnit.MILLISECONDS));
    } finally {
      validationCollector.logEvents("deserialize: " + clazz.getName());
    }
    return object;
  }

  public static <T> JAXBElement<T> deserializeNoKnowRootElement(final String content, final Class<?> clazz) throws JAXBException {
    final JAXBElement<T> object;
    final JaxbValidationEventCollector validationCollector = new JaxbValidationEventCollector();
    try {
      final Unmarshaller jaxbUnmarshaller = getJaxbContext(clazz).createUnmarshaller();
      jaxbUnmarshaller.setEventHandler(validationCollector);
      final Stopwatch s = Stopwatch.createStarted();
      try (StringReader r = new StringReader(content)) {
        object = (JAXBElement<T>) jaxbUnmarshaller.unmarshal(new StreamSource(r), clazz);
      }
      log.debug("deserialize time: " + s.elapsed(TimeUnit.MILLISECONDS));
    } finally {
      validationCollector.logEvents("deserialize: " + clazz.getName());
    }
    return object;
  }


  private static class JaxbValidationEventCollector extends ValidationEventCollector {

    @Override
    public boolean handleEvent(final ValidationEvent event) {
      super.handleEvent(event);
      return true;
    }

    public void logEvents(final String methodName) {
      if (hasEvents()) {
        log.warn(" JAXB-warning in method: " + methodName);
        for (final ValidationEvent event : getEvents()) {
          final String msg = event.toString();
          log.warn(msg);
        }
      }
    }
  }
}
