/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.util;

import de.xoev.xta.client.exception.ClientInitializationException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.HashMap;
import java.util.Map;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.StringUtils;

/**
 * Hilfsklasse für KeyStore-Handling.
 */
@SuppressWarnings("checkstyle:HideUtilityClassConstructor")
@NoArgsConstructor(access = AccessLevel.NONE)
public final class KeyStoreUtil {

  private static Map<String, String> mapping = new HashMap<>();

  static {
    mapping.put("p12", "PKCS12");
    mapping.put("jceks", "JCEKS");
    mapping.put("jks", "JKS");
  }

  /**
   * Läd und liefert einen KeyStrore.
   * <p>
   * Unterstützte Formate (bisher): .p12 (Typ: PKCS12) und .jceks (Typ: JCEKS)
   *
   * @param path       der Dateienpfad zur Store-Datei
   * @param passPhrase Passwort des Stores
   * @return der durch den Pfad referenzierte KeyStore
   * @throws ClientInitializationException bei zum Beispiel nicht unterstützen Typen oder wenn die KeyStore-Datei nicht gefunden
   *                                       wurde
   */
  public static KeyStore load(final String path, final String passPhrase) throws ClientInitializationException {

    try {
      final KeyStore keyStore = KeyStore.getInstance(getType(path));
      try (final InputStream is = new FileInputStream(path)) {
        keyStore.load(is, passPhrase.toCharArray());
      } catch (FileNotFoundException e) {
        // relativ zum Classpath: Fallback for testing
        try (final InputStream is = KeyStoreUtil.class.getClassLoader().getResourceAsStream(path)) {
          keyStore.load(is, passPhrase.toCharArray());
        }
      }
      return keyStore;
    } catch (IOException | NoSuchAlgorithmException | CertificateException | KeyStoreException e) {
      throw new ClientInitializationException("Exception in createSSLContext: " + e.getMessage(), e);
    }
  }

  private static String getType(final String path) throws ClientInitializationException {
    if (!StringUtils.isEmpty(path)) {
      String key = path.substring(path.lastIndexOf(".") + 1).toUpperCase();
      return mapping.get(key.toLowerCase());
    }
    return KeyStore.getDefaultType();
  }

  public static SecretKey generateSecretKey() throws NoSuchAlgorithmException {
    KeyGenerator keygen = KeyGenerator.getInstance("AES");
    keygen.init(128);
    return keygen.generateKey();
  }
}
