/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.util;

import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.exception.ClientInitializationException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Map;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import jodd.bean.BeanUtil;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Utility-Klasse für XML-Funktionen.
 */
@SuppressWarnings("checkstyle:HideUtilityClassConstructor")
@NoArgsConstructor(access = AccessLevel.NONE)
public final class XmlUtil {

  public static List<Date> fromXmlGregorianCalender(final List<XMLGregorianCalendar> xmlGregorianCalendarList) {
    final List<Date> dateList = new ArrayList<>();
    for (final XMLGregorianCalendar xmlGregorianCalendar : xmlGregorianCalendarList) {
      dateList.add(fromXmlGregorianCalender(xmlGregorianCalendar));
    }

    return dateList;
  }

  /**
   * Transformiert einen XMLGregorianCalendar in Date.
   *
   * @param xmlGregorianCalendar XMLGregorianCalendar zu tranformierende XMLGregorianCalendar-Instanz
   * @return übergebenden XMLGregorianCalendar als Date
   * @see {@link #toXmlGregorianCalender(Date)}
   */
  public static Date fromXmlGregorianCalender(final XMLGregorianCalendar xmlGregorianCalendar) {
    return (xmlGregorianCalendar == null ? null : xmlGregorianCalendar.toGregorianCalendar().getTime());
  }

  /**
   * Transformiert ein Date zu XMLGregorianCalendar.
   *
   * @param date zu tranformierende Date-Instanz
   * @return übergebendendes Date als XMLGregorianCalendar
   * @see {@link #fromXmlGregorianCalender(XMLGregorianCalendar)}
   */
  public static XMLGregorianCalendar toXmlGregorianCalender(final Date date) throws ClientException {
    try {
      if (date != null) {
        final GregorianCalendar calStart = new GregorianCalendar();
        calStart.setTime(date);
        return DatatypeFactory.newInstance().newXMLGregorianCalendar(calStart);
      } else {
        return null;
      }
    } catch (DatatypeConfigurationException e) {
      throw new ClientException("Exception in toXmlGregorianCalender: " + e.getMessage(), e);
    }
  }

  public static Document toNsAwareDocument(final byte[] xml) throws ClientException {
    try {
      final DocumentBuilderFactory s = DocumentBuilderFactory.newInstance();
      s.setNamespaceAware(true);
      final DocumentBuilder db = s.newDocumentBuilder();
      final Document document = db.parse(new ByteArrayInputStream(xml));
      return document;
    } catch (ParserConfigurationException | SAXException | IOException e) {
      throw new ClientException("Exception in toNsAwareDocument: " + e.getMessage(), e);
    }
  }

  public static String toString(final Document source) throws TransformerException {
    final StringWriter writer = new StringWriter();
    final StreamResult result = new StreamResult(writer);
    final TransformerFactory tf = TransformerFactory.newInstance();
    final Transformer transformer = tf.newTransformer();
    transformer.transform(new DOMSource(source), result);
    return writer.toString();
  }

  public static String toString(Node node) throws TransformerException {
    StringWriter sw = new StringWriter();
    Transformer t = TransformerFactory.newInstance().newTransformer();
    t.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
    t.setOutputProperty(OutputKeys.INDENT, "yes");
    t.transform(new DOMSource(node), new StreamResult(sw));
    return sw.toString();
  }


  /**
   * Hilfsmethode zum Übertragen von XML-Daten (aus eine Fachnachricht mit Nachrichtenkopf) in die entsprechenden Instanz.
   *
   * @param xml      die (XML-)Fachnachricht (Quelle)
   * @param target   die Objektinstanz (Ziel)
   * @param mappings das Mapping, also die Vorgaben welches XML-Element gehört in welche Eigenschaft der Objektinstanz.
   * @param <T>      typ der Objektinstanz
   * @return die Objektinstanz, gefüllt mit Werten aus der Fachnachricht
   * @throws ClientInitializationException
   */
  public static <T> T transferMessageDataIntoBean(String xml, T target, Map<String, String> mappings)
      throws ClientInitializationException {

    try {
      final DocumentBuilderFactory s = DocumentBuilderFactory.newInstance();

      s.setNamespaceAware(false);

      final DocumentBuilder db = s.newDocumentBuilder();

      InputSource is = new InputSource(new StringReader(xml));
      final Document doc = db.parse(is);

      XPathFactory xf = XPathFactory.newInstance();
      XPath xpath = xf.newXPath();

      for (Map.Entry<String, String> e : mappings.entrySet()) {
        final XPathExpression expression = xpath.compile(e.getValue());
        final Class<?> clazz = BeanUtil.declared.getPropertyType(target, e.getKey());
        if (clazz.isAssignableFrom(Boolean.class)) {
          Boolean value = (Boolean) expression.evaluate(doc, XPathConstants.BOOLEAN);
          BeanUtil.declared.setProperty(target, e.getKey(), value);
        } else if (clazz.isAssignableFrom(Date.class)) {
          String value = expression.evaluate(doc);
          Instant instant = Instant.parse(value);
          Date d = java.util.Date.from(instant);
          BeanUtil.declared.setProperty(target, e.getKey(), d);
        } else {
          String value = expression.evaluate(doc);
          if (StringUtils.isNotBlank(value)) {
            BeanUtil.declared.setProperty(target, e.getKey(), value);
          }
        }
      }
      return target;

    } catch (SAXException | ParserConfigurationException | IOException | XPathExpressionException e) {
      throw new ClientInitializationException("Exception in transferMessageDataIntoBean: " + e.getMessage(), e);
    }
  }
}
