/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

import de.xoev.xta.client.codes.BusinessScenario;
import de.xoev.xta.client.config.XtaClientSettings;
import de.xoev.xta.client.encryption.SampleEncryptionService;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.exception.ClientInitializationException;
import de.xoev.xta.client.mapping.MessageMapping;
import de.xoev.xta.client.model.DeliveryOptions;
import de.xoev.xta.client.model.DeliveryStatus;
import de.xoev.xta.client.model.GetMessageResult;
import de.xoev.xta.client.model.Identifier;
import de.xoev.xta.client.model.Message;
import de.xoev.xta.client.model.MessageFilter;
import de.xoev.xta.client.model.MessageTransportReport;
import java.time.Instant;
import java.util.Date;
import java.util.List;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

/**
 * Testcase - Hinweise:
 * <p>
 * Zum Ausführen aller Tests werden Dateien im Filesystem benötigt:
 * <ul>
 *     <li>Keystore für SampleEncryptionService (siehe KEYSTORE_PATH und SampleEncryptionServiceTest)</li>
 *     <li>config/messagemapping.properties</li>
 *     <li>config/xtaclient_author.properties mit gesetzten Einstellungen</li>
 *     <li>config/xtaclient_reader.properties mit gesetzten Einstellungen</li>
 *     <li>je nach Einstellungen Keystore und Zertifikate in certs/..</li>
 * </ul>
 * <p>
 * Die Datein sind auch bestandteil des vorliegenden JAR's
 * <p>
 * Desweiteren benötigt man 2 aktive Accounts beim Nachrichtenvermittler.
 */
@Ignore("Einrichtung notwendig")
public class BrokerTests extends BaseTest {

  final BusinessScenario bs = BusinessScenario.XDOMEAREG_DATA;
  final String serviceName = "http://www.osci.de/xdomea23/xdomea23Reg2Reg.wsdl";

  private XtaClientSettings xtaClientSettings_author;
  private XtaClientSettings xtaClientSettings_reader;

  @Before
  public void initializeSettings() throws ClientInitializationException {
    xtaClientSettings_author = new XtaClientSettings(
        BaseTest.class.getClassLoader().getResourceAsStream("config/xtaclient_author.properties"));

    xtaClientSettings_reader = new XtaClientSettings(
        BaseTest.class.getClassLoader().getResourceAsStream("config/xtaclient_reader.properties"));
  }

  @Test
  public void sendMessage_normalCase() throws Exception {

    // Nachricht einlesen
    final Message message = createTestMessage();

    // die Identität aus Einstellungen erzeugen (Achtung Autor = Sender = Leser)
    Identifier authorId = Identifier.from(xtaClientSettings_author);

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient client = initializeClient(xtaClientSettings_author);

    // DeliveryOptions erzeugen - BusinessScenario: XDOMEAREG_DATA, Service: http://www.osci.de/xdomea23/xdomea23Reg2Reg.wsdl
    DeliveryOptions deliveryOptions = new DeliveryOptions(serviceName, BusinessScenario.XDOMEAREG_DATA);

    // senden der Nachricht (an sich selbst)
    final MessageTransportReport transportReport = client.sendMessage(authorId, deliveryOptions, message);

    assertThat(transportReport.getMessageId()).isNotNull();
  }

  @Test
  @Ignore //Flag isTestMsg ist nicht gesetzt.
  public void sendMessage_testMessage() throws Exception {

    // Nachricht einlesen
    final Message message = createTestMessage();

    // die Identität aus Einstellungen erzeugen (Achtung Autor = Sender = Leser)
    Identifier authorId = Identifier.from(xtaClientSettings_author);

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient client = initializeClient(xtaClientSettings_author);

    // DeliveryOptions erzeugen - als TESTNACHRICHT markiert, BusinessScenario: XDOMEAREG_DATA, Service: http://www.osci.de/xdomea23/xdomea23Reg2Reg.wsdl
    DeliveryOptions deliveryOptions = new DeliveryOptions(serviceName, bs).asTestMessage();

    // senden der Nachricht (an sich selbst)
    final MessageTransportReport transportReport = client.sendMessage(authorId, deliveryOptions, message);

    // Test
    assertThat(transportReport.getMessageId()).isNotNull();
    assertThat(transportReport.getXtaTransportReport().getMessageMetaData().isTestMsg()).isTrue();
  }


  @Test
  public void sendMessage_unkownService() throws Exception {

    // Nachricht einlesen
    final Message message = createTestMessage();

    // die Identität aus Einstellungen erzeugen (Achtung Autor = Leser)
    Identifier authorId = Identifier.from(xtaClientSettings_author);

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient client = initializeClient(xtaClientSettings_author);

    // DeliveryOptions erzeugen - mit unbekannten Service
    DeliveryOptions deliveryOptions = new DeliveryOptions("Unbekannter-Service:1.0", bs);

    assertThatThrownBy(() -> {
      //  - Validierungsfehler werden hier unterdrückt
      // senden der Nachricht (an sich selbst)
      client.sendMessage(authorId, deliveryOptions, message);
    }).isInstanceOf(ClientException.class).hasMessageContaining("is not available");
  }


  @Test
  public void sendMessage_withIdAndOptionsFromMessage()
      throws Exception {

    // Nachricht einlesen
    final Message message = createTestMessage();
    final MessageMapping msgMapper = new MessageMapping(message,
        BrokerTests.class.getClassLoader().getResourceAsStream("config/messagemapping.properties"));

    // die AUTHOR-Identität aus Fachnaricht erzeugen
    Identifier authorId = msgMapper.getIdentifier(Identifier.CommunicationRole.AUTHOR);

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient client = initializeClient(authorId, xtaClientSettings_author);

    // die Empfänger-Identität aus Fachnachricht erzeugen
    Identifier readerId = msgMapper.getIdentifier(Identifier.CommunicationRole.READER);

    // DeliveryOptions aus Fachnaricht erzeugen
    DeliveryOptions deliveryOptions = msgMapper.getDeliveryOptions().withCheckDeliveryStatus(DeliveryStatus.RECEIVED_IN_MSGBOX);

    // senden der Nachricht (an receiver)
    final MessageTransportReport transportReport = client.sendMessage(readerId, deliveryOptions, message);

    // Test
    assertThat(transportReport.getMessageId()).isNotNull();
  }


  @Test
  public void checkMailBox_afterSending() throws Exception {

    // die Identität aus Einstellungen erzeugen (Achtung Autor = Leser)
    Identifier authorId = Identifier.from(xtaClientSettings_author);

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient client = initializeClient(xtaClientSettings_author);

    // erzeugen und senden der Nachricht (an receiver) - um diese dann im nächsten Schritt zu finden
    final Message message = createTestMessage();
    final DeliveryOptions deliveryOptions = new DeliveryOptions(serviceName, bs)
        .withCheckDeliveryStatus(DeliveryStatus.RECEIVED_IN_MSGBOX);
    final MessageTransportReport report = client.sendMessage(authorId, deliveryOptions, message);

    // -- Nachrichtenselktor erzeugen --
    // - alle Nachrichten die ab "Sysdate - 10 minuten" angekommen sind
    final Date sendDate = Date.from(Instant.now().minusSeconds(600)); // 10 minuten

    // - nur neue Nachrichten mit der erwartetenden MessageId
    MessageFilter messageFilter = new MessageFilter().newEntries().withCount(500).withTimeFrom(sendDate)
        .withMessageIds(report.getMessageId());
    List<MessageTransportReport> messageTransportReportList = client.getMessageTransportReportList(messageFilter);

    // Test - es muss mindestens eine neue Nachricht vorhanden sein
    assertThat(messageTransportReportList.size()).isGreaterThanOrEqualTo(1);
  }


  @Test
  public void checkMailBox_noMsgAvailable() throws Exception {

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient client = initializeClient(xtaClientSettings_author);

    // -- Nachrichtenselktor erzeugen --
    // - alle Nachrichten die ab "Sysdate + 10 minuten" angekommen sind - also Zukunft
    final Date sendDate = Date.from(Instant.now().plusSeconds(600)); // 10 minuten

    // - nur die erste neue Nachricht -> da das Datum in der Zukunft liegt sollte das Ergebnis so immer 0 sein
    MessageFilter messageFilter = new MessageFilter().newEntries().withCount(1).withTimeFrom(sendDate);
    List<MessageTransportReport> messageTransportReportList = client.getMessageTransportReportList(messageFilter);

    // Test - es dürfen keine Nachrichten in der Mailbox sein
    assertThat(messageTransportReportList.size()).isEqualTo(0);
  }


  @Test
  public void roundtripTest_withSenderAndReceiver() throws Exception {

    // XTA-Clients mit den Identitäten erzeugen
    final XtaClient authorClient = initializeClient(xtaClientSettings_author);
    final XtaClient readerClient = initializeClient(xtaClientSettings_reader);

    final Identifier readerId = new Identifier("xdo:11009992", "xoev");

    final DeliveryOptions deliveryOptions = new DeliveryOptions(serviceName, bs)
        .withCheckDeliveryStatus(DeliveryStatus.RECEIVED_IN_MSGBOX);

    // Nachricht erzeugen und mit dem Sender-Client an Empfänger senden
    final Message message = createTestMessage();
    final MessageTransportReport transportReport = authorClient.sendMessage(readerId, deliveryOptions, message);
    final String messageIdSend = transportReport.getMessageId();

    // Postfachabfrage am Empfänger
    final MessageFilter messageFilter = new MessageFilter().newEntries().withCount(100)
        .withTimeFrom(Date.from(Instant.now().minusSeconds(60)));
    List<MessageTransportReport> messageTransportReportList = readerClient.getMessageTransportReportList(messageFilter);

    // Test - es muss mindestens eine neue Nachricht vorhanden sein
    assertThat(messageTransportReportList.size()).isGreaterThan(0);

    // die vom Sender gesendete Nachricht sollte gleich einer Nachricht im Postfach des Empfängers sein
    final String messageIdReceive = getAndCheckMessageInMessageTransportList(messageIdSend, messageTransportReportList);

    // Test - sender-nachrichten-ID == empfänger-nachrichten-Id
    // (!) klappt nur wenn Empfänger und Sender beim selben Broker "gehostet" werden
    assertThat(messageIdReceive).isEqualTo(messageIdSend);

    // ganze Nachricht aus Postfach abholen
    final GetMessageResult getMessageResult = readerClient.getMessage(messageIdReceive);

    // Test - gesendete Nachricht == empfangende Nachricht
    assertThat(getMessageResult.getMessage().getData()).isEqualTo(message.getData());

    // -------------------------------------------------------------------
    // Etwas mit der Nachricht machen: getMessageResult.getMessage().getData()
    // -------------------------------------------------------------------

    // nach Auswerten/Bearbeiten/Etwas-mit-der-nachricht-machen - wird die nachricht quttiert "geschlossen"

    // receiver.close(messageIdReceive, getMessageResult.getRequestHandle());
    readerClient.close(messageIdReceive, messageIdReceive);

    //  Auskommentiert weil Receipt vom Broker nicht korrekt gesetzt ist und daher nicht der korrekte Delivery-Status ermittelt werden kann.
    //  - sender: prüfen ob die Nachricht vom Empfänger abgeholt und quittiert wurde
    // checkDeliveryStatus(authorClient, messageIdSend, DeliveryStatus.RECEIPT_FROM_MSGBOX);
  }

  @Test
  @Ignore // führt zu Fehler. Daher erstmal auskommentiert.
  public void sendMessage_crypt() throws Exception {

    /// die Identität aus Einstellungen erzeugen (Achtung Autor = Leser)
    Identifier authorId = Identifier.from(xtaClientSettings_author);

    // XTA-Client mit Sender-Identität erzeugen
    XtaClient authorClient = new XtaClient(XtaClientConfig.create(xtaClientSettings_author)
        .withSoapLogging().withEncryptionServices(new SampleEncryptionService()).build());

    final Message message = createTestMessage();

    // DeliveryOptions erzeugen - als TESTNACHRICHT markiert, BusinessScenario: XDOMEAREG_DATA, Service: http://www.osci.de/xdomea23/xdomea23Reg2Reg.wsdl
    final DeliveryOptions deliveryOptions = new DeliveryOptions(serviceName, bs)
        .withCheckDeliveryStatus(DeliveryStatus.RECEIVED_IN_MSGBOX)
        .withEncryptionServiceProfile("SAMPLE");

    final MessageTransportReport transportReport = authorClient.sendMessage(authorId, deliveryOptions, message);

    // Nachricht holen
    final GetMessageResult result = authorClient.getMessage(transportReport.getMessageId());
    assertThat(result.getMessage().getFilename()).isEqualTo(message.getFilename());

  }
}
