/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.encryption;

import java.security.Key;
import java.security.PublicKey;
import java.util.List;
import javax.xml.namespace.QName;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;
import org.apache.xml.security.Init;
import org.apache.xml.security.encryption.EncryptedData;
import org.apache.xml.security.encryption.EncryptedKey;
import org.apache.xml.security.encryption.XMLCipher;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.stax.ext.XMLSecurityConstants;
import org.assertj.core.api.Assertions;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Some utility methods for encrypting/decrypting documents
 */
public final class EncryptionUtils {

  static {
    Init.init();
  }

  private EncryptionUtils() {
    // complete
  }

  /**
   * Encrypt the document using the DOM API of Apache Santuario - XML Security for Java. It encrypts a list of QNames that it
   * finds in the Document via XPath. If a wrappingKey is supplied, this is used to encrypt the encryptingKey + place it in an
   * EncryptedKey structure.
   */
  public static void encryptUsingDOM(
      Document document,
      List<QName> namesToEncrypt,
      String algorithm,
      Key encryptingKey,
      String keyTransportAlgorithm,
      PublicKey wrappingKey,
      boolean content
  ) throws Exception {
    XMLCipher cipher = XMLCipher.getInstance(algorithm);
    cipher.init(XMLCipher.ENCRYPT_MODE, encryptingKey);

    if (wrappingKey != null) {
      XMLCipher newCipher = XMLCipher.getInstance(keyTransportAlgorithm);
      newCipher.init(XMLCipher.WRAP_MODE, wrappingKey);

      EncryptedKey encryptedKey = newCipher.encryptKey(document, encryptingKey);
      // Create a KeyInfo for the EncryptedKey
      KeyInfo encryptedKeyKeyInfo = encryptedKey.getKeyInfo();
      if (encryptedKeyKeyInfo == null) {
        encryptedKeyKeyInfo = new KeyInfo(document);
        encryptedKeyKeyInfo.getElement().setAttributeNS(
            "http://www.w3.org/2000/xmlns/", "xmlns:dsig", "http://www.w3.org/2000/09/xmldsig#"
        );
        encryptedKey.setKeyInfo(encryptedKeyKeyInfo);
      }
      encryptedKeyKeyInfo.add(wrappingKey);

      // Create a KeyInfo for the EncryptedData
      EncryptedData builder = cipher.getEncryptedData();
      KeyInfo builderKeyInfo = builder.getKeyInfo();
      if (builderKeyInfo == null) {
        builderKeyInfo = new KeyInfo(document);
        builderKeyInfo.getElement().setAttributeNS(
            "http://www.w3.org/2000/xmlns/", "xmlns:dsig", "http://www.w3.org/2000/09/xmldsig#"
        );
        builder.setKeyInfo(builderKeyInfo);
      }

      builderKeyInfo.add(encryptedKey);
    }

    XPathFactory xpf = XPathFactory.newInstance();
    XPath xpath = xpf.newXPath();
    xpath.setNamespaceContext(new DSNamespaceContext());

    for (QName nameToEncrypt : namesToEncrypt) {
      String expression = "//*[local-name()='" + nameToEncrypt.getLocalPart() + "']";
      NodeList elementsToEncrypt =
          (NodeList) xpath.evaluate(expression, document, XPathConstants.NODESET);
      for (int i = 0; i < elementsToEncrypt.getLength(); i++) {
        Element elementToEncrypt = (Element) elementsToEncrypt.item(i);
        Assertions.assertThat(elementToEncrypt).isNotNull();
        document = cipher.doFinal(document, elementToEncrypt, content);
      }
    }

    String expression = "//xenc:EncryptedData[1]";
    Element encElement =
        (Element) xpath.evaluate(expression, document, XPathConstants.NODE);
    Assertions.assertThat(encElement).isNotNull();
  }

  /**
   * Decrypt the document using the DOM API of Apache Santuario - XML Security for Java.
   */
  public static void decryptUsingDOM(
      Document document,
      String algorithm,
      Key privateKey
  ) throws Exception {
    XMLCipher cipher = XMLCipher.getInstance(algorithm);
    cipher.init(XMLCipher.DECRYPT_MODE, null);
    cipher.setKEK(privateKey);

    NodeList nodeList = document.getElementsByTagNameNS(
        XMLSecurityConstants.TAG_xenc_EncryptedData.getNamespaceURI(),
        XMLSecurityConstants.TAG_xenc_EncryptedData.getLocalPart()
    );

    for (int i = 0; i < nodeList.getLength(); i++) {
      Element ee = (Element) nodeList.item(i);
      cipher.doFinal(document, ee);
    }
  }

}