/**
 *     XTA-Client Java Library
 *     Copyright (C) 2021 Koordinierungsstelle für IT-Standards (KoSIT)
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.xoev.xta.client.encryption;

import de.xoev.xta.client.encyrption.EncryptionService;
import de.xoev.xta.client.exception.ClientException;
import de.xoev.xta.client.util.JaxbUtil;
import de.xoev.xta.client.util.KeyStoreUtil;
import de.xoev.xta.client.util.XmlUtil;
import de.xoev.transport.xta._211.GenericContentContainer.ContentContainer;
import java.io.ByteArrayInputStream;
import java.security.Key;
import java.security.KeyStore;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.List;
import javax.crypto.SecretKey;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;

import org.apache.commons.io.IOUtils;
import org.apache.xml.security.utils.XMLUtils;
import org.w3._2001._04.xmlenc_.EncryptedDataType;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * Beispielimplementierung eines EncyrptionService (verwendet XML-Santuario)
 */
public class SampleEncryptionService implements EncryptionService {

  private static final String TRANSPORT_ALGORITHM = "http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p";
  private static final String ALGORITHM = "http://www.w3.org/2001/04/xmlenc#aes256-cbc";

  private static final String KEYSTORE_PATH = "certs/sampleencytionservice.jks";
  private static final String KEYSTORE_PASSWORD = "...";
  private static final String CERT_ALIAS = "test";
  private static final String CERT_PRIVATEKEY_PASSWORD = "..";

  @Override
  public String getServiceProfilname() {
    return "SAMPLE";
  }

  @Override
  public EncryptedDataType encrypt(final ContentContainer contentContainer) throws ClientException {
    try {
      final JAXBElement<ContentContainer> jaxbElement = new JAXBElement<>(new QName("", "contentContainer"),
          ContentContainer.class, contentContainer);

      final Document document = XMLUtils.createDocumentBuilder(false)
          .parse(IOUtils.toInputStream(JaxbUtil.serialize(jaxbElement, ContentContainer.class), "UTF-8"));

      final KeyStore keyStore = KeyStoreUtil.load(KEYSTORE_PATH, KEYSTORE_PASSWORD);
      final X509Certificate cert = (X509Certificate) keyStore.getCertificate(CERT_ALIAS);
      final SecretKey secretKey = KeyStoreUtil.generateSecretKey();

      final List<QName> namesToEncrypt = new ArrayList<QName>();
      namesToEncrypt.add(new QName("", "contentContainer"));
      EncryptionUtils.encryptUsingDOM(
          document, namesToEncrypt, ALGORITHM, secretKey,
          TRANSPORT_ALGORITHM, cert.getPublicKey(), false
      );
      final String xml = XmlUtil.toString(document);
      return (EncryptedDataType) ((JAXBElement) JaxbUtil.deserialize(xml, EncryptedDataType.class)).getValue();
    } catch (final Exception e) {
      throw new ClientException(e.getMessage(), e);
    }
  }

  @Override
  public ContentContainer decyrpt(final EncryptedDataType encryptedDataType) throws ClientException {
    try {
      final JAXBElement<EncryptedDataType> jaxbElement = new JAXBElement<>(
          new QName("http://www.w3.org/2001/04/xmlenc#", "EncryptedDataType"),
          EncryptedDataType.class, encryptedDataType);
      final String xml = JaxbUtil.serialize(jaxbElement, EncryptedDataType.class)
          .replaceAll("EncryptedDataType", "EncryptedData");

      final Document document = XMLUtils.createDocumentBuilder(false).parse(new ByteArrayInputStream(xml.getBytes()));

      final KeyStore keyStore = KeyStoreUtil.load(KEYSTORE_PATH, KEYSTORE_PASSWORD);
      final Key privateKey = keyStore.getKey(CERT_ALIAS, CERT_PRIVATEKEY_PASSWORD.toCharArray());

      EncryptionUtils.decryptUsingDOM(document,
          ALGORITHM, privateKey);

      final NodeList nodeList = document.getElementsByTagNameNS("", "contentContainer");
      if (nodeList.getLength() == 0) {
        throw new ClientException("Inhalt kann nicht entschlüsselt werden");
      }

      final JAXBElement<ContentContainer> container = JaxbUtil
          .deserializeNoKnowRootElement(XmlUtil.toString(nodeList.item(0)), ContentContainer.class);
      return container.getValue();

    } catch (final Exception e) {
      throw new ClientException(e.getMessage(), e);
    }
  }
}
